#include "StdAfx.h"
#include "ArchiveScanner.h"
#include <io.h>
#include "ArchiveFactory.h"
#include <algorithm>
#include "SunParser.h"
#include "mygl.h"
#include "FileHandler.h"

#include <sys/types.h>
#include <sys/stat.h>

/*
 * The archive scanner is used to find stuff in archives that are needed before building the virtual
 * filesystem. This currently includes maps and mods. It uses caching to speed up the process.
 *
 * It only retrieves info that is used in an initial listing.. For detailed info when selecting a map
 * for example the more specialised parsers will be used. (mapping one archive when selecting a map
 * is not slow, but mapping them all every time to make the list is)
 */

CArchiveScanner* archiveScanner = NULL;

CArchiveScanner::CArchiveScanner(void) :
	isDirty(false)
{
}

CArchiveScanner::~CArchiveScanner(void)
{
	if (isDirty)
		WriteCacheData();
}

CArchiveScanner::ModData CArchiveScanner::GetModData(CSunParser* p, const string& section)
{
	ModData md;
	md.name = "";

	if (!p->SectionExist(section)) 
		return md;

	md.name = p->SGetValueDef("", (section + "\\Name").c_str());
	md.description = p->SGetValueDef("", (section + "\\Description").c_str());
	md.modType = atoi(p->SGetValueDef("0", (section + "\\ModType").c_str()).c_str());

	int numDep = atoi(p->SGetValueDef("0", (section + "\\NumDependencies").c_str()).c_str());
	for (int dep = 0; dep < numDep; ++dep) {
		char key[100];
		sprintf(key, "%s\\Depend%d", section.c_str(), dep);
		md.dependencies.push_back(p->SGetValueDef("", key));
	}

	return md;
}

void CArchiveScanner::Scan(const string& curPath)
{
	struct _finddata_t files;
	intptr_t handle;
	int moreFiles;

	string curPattern = curPath;
	curPattern += "/*.*";

	char t[500];
	sprintf(t, "Scanning archives in %s", curPath.c_str());
	//PrintLoadMsg(t);

	isDirty = true;

	handle = _findfirst(curPattern.c_str(), &files);
	moreFiles = (int)handle;
	while (moreFiles != -1) {

		//printf("looking at %s\n", files.name);

		// Is it a directory?
		if ((files.attrib & 16) != 0) {

			// Avoid the special directories
			if (files.name[0] != '.') {
				string newPath = curPath + "/" + files.name;
				Scan(newPath);
			}
		}

		// Or maybe an archive we should look into?
		else if (CArchiveFactory::IsArchive(files.name)) {

			// Determine whether to rely on the cached info or not
			bool cached = false;
			string lcname = files.name;
			transform(lcname.begin(), lcname.end(), lcname.begin(), (int (*)(int))tolower);
			string fullName = curPath + "/" + files.name;

			map<string, ArchiveInfo>::iterator aii = archiveInfo.find(lcname);
			if (aii != archiveInfo.end()) {
				struct _stat si;
				if (_stat(fullName.c_str(), &si) == 0) {
					if (si.st_mtime == aii->second.modified) {
						cached = true;
						aii->second.updated = true;
					}
				}

				// If we are here, we could have invalid info in the cache
				if (!cached) {
					archiveInfo.erase(aii);
				}
			}

			// Time to parse the info we are interested in
			if (!cached) {
				CArchiveBase* ar = CArchiveFactory::OpenArchive(fullName);
				if (ar) {
					int cur;
					string name;
					int size;
					ArchiveInfo ai;

					cur = ar->FindFiles(0, &name, &size);
					while (cur != 0) {
						//printf("found %s %d\n", name.c_str(), size);
						
						string ext = name.substr(name.find_last_of('.') + 1);
						transform(ext.begin(), ext.end(), ext.begin(), (int (*)(int))tolower);

						// only accept new format maps
						if (ext == "smf") {
							MapData md;
							if (name.find_last_of('\\') == string::npos) {
								md.name = name;
								md.virtualPath = "\\";
							}
							else {
								md.name = name.substr(name.find_last_of('\\') + 1);
								//md.name = md.name.substr(0, md.name.find_last_of('.'));
								md.virtualPath = name.substr(0, name.find_last_of('\\') + 1);	// include the backslash
							}
							ai.mapData.push_back(md);
						}

						if (name == "modinfo.tdf") {
							CSunParser *p = new CSunParser();
							int fh = ar->OpenFile(name);
							if (fh) {
								int fsize = ar->FileSize(fh);
								void* buf = malloc(fsize);
								ar->ReadFile(fh, buf, fsize);
								ar->CloseFile(fh);
								p->LoadBuffer((char*)buf, fsize);	
								free(buf);
								
								ai.modData = GetModData(p, "mod");
							}
							delete p;

						}

						cur = ar->FindFiles(cur, &name, &size);
					}

					ai.path = curPath;
					struct _stat si;
					_stat(fullName.c_str(), &si);
					ai.modified = si.st_mtime;
					ai.origName = files.name;
					ai.updated = true;

					archiveInfo[lcname] = ai;

					delete ar;
				}
			}
		}

		moreFiles = _findnext(handle, &files);
	}
	_findclose(handle);
}

void CArchiveScanner::ReadCacheData()
{
	// check if the file exists first
	CFileHandler file("archivecache.txt");
	if (!file.FileExists())
		return;

	CSunParser p;
	p.LoadFile("archivecache.txt");

	int numArs = atoi(p.SGetValueDef("0", "archivecache\\numarchives").c_str());
	for (int i = 0; i < numArs; ++i) {
		ArchiveInfo ai;
		char keyb[100];
		sprintf(keyb, "ArchiveCache\\Archive%d\\", i);
		string key = keyb;

		ai.origName = p.SGetValueDef("", key + "Name");
		ai.path = p.SGetValueDef("", key + "Path");
		ai.modified = atol(p.SGetValueDef("", key + "Modified").c_str());
		ai.updated = false;

		int numMaps = atoi(p.SGetValueDef("0", key + "NumMaps").c_str());
		for (int m = 0; m < numMaps; ++m) {
			char mapb[100];
			sprintf(mapb, "%sMap%d\\", key.c_str(), m);
			string map = mapb;

			MapData md;
			md.name = p.SGetValueDef("", map + "Name");
			md.virtualPath = p.SGetValueDef("", map + "VirtualPath");

			ai.mapData.push_back(md);
		}

		if (p.SectionExist(key + "Mod")) {
			ai.modData = GetModData(&p, key + "Mod");
		}

		string lcname = ai.origName;
		transform(lcname.begin(), lcname.end(), lcname.begin(), (int (*)(int))tolower);

		archiveInfo[lcname] = ai;
	}

	isDirty = false;
}

void CArchiveScanner::WriteCacheData()
{
	if (!isDirty)
		return;

	FILE* out = fopen("archivecache.txt", "wt");
	//FILE* out = stdout;
	if (!out)
		return;
	
	// First delete all outdated information
	for (map<string, ArchiveInfo>::iterator i = archiveInfo.begin(); i != archiveInfo.end(); ) {
		map<string, ArchiveInfo>::iterator next = i;
		next++;
		if (!i->second.updated) {
			archiveInfo.erase(i);
		}
		i = next;
	}

	fprintf(out, "[ARCHIVECACHE]\n{\n");
	fprintf(out, "\tNumArchives=%d;\n", archiveInfo.size());
	int cur = 0;
	for (map<string, ArchiveInfo>::iterator i = archiveInfo.begin(); i != archiveInfo.end(); ++i) {
		fprintf(out, "\t[ARCHIVE%d]\n\t{\n", cur);
		fprintf(out, "\t\tName=%s;\n", i->second.origName.c_str());
		fprintf(out, "\t\tPath=%s;\n", i->second.path.c_str());
		fprintf(out, "\t\tModified=%u;\n", i->second.modified);

		fprintf(out, "\t\tNumMaps=%d;\n", i->second.mapData.size());
		int curmap = 0;
		for (vector<MapData>::iterator mi = i->second.mapData.begin(); mi != i->second.mapData.end(); ++mi) {
			fprintf(out, "\t\t[MAP%d]\n\t\t{\n", curmap);
			//WriteData(out, *mi);
			fprintf(out, "\t\t\tName=%s;\n", (*mi).name.c_str());
			fprintf(out, "\t\t\tVirtualPath=%s;\n", (*mi).virtualPath.c_str());
			//fprintf(out, "\t\t\tDescription=%s;\n", (*mi).description.c_str());
			fprintf(out, "\t\t}\n");
		}

		// Any mod info? or just a map archive?
		if (i->second.modData.name != "") {
			fprintf(out, "\t\t[MOD]\n\t\t{\n");
			fprintf(out, "\t\t\tName=%s;\n", i->second.modData.name.c_str());
			fprintf(out, "\t\t\tDescription=%s;\n", i->second.modData.description.c_str());
			fprintf(out, "\t\t\tModType=%d;\n", i->second.modData.modType);
			fprintf(out, "\t\t\tNumDependencies=%d;\n", i->second.modData.dependencies.size());

			int curdep = 0;
			for (vector<string>::iterator dep = i->second.modData.dependencies.begin(); dep != i->second.modData.dependencies.end(); ++dep) {
				fprintf(out, "\t\t\tDepend%d=%s;\n", curdep, (*dep).c_str());
				curdep++;
			}

			fprintf(out, "\t\t}\n");
		}

		fprintf(out, "\t}\n");
		cur++;
	}
	fprintf(out, "}\n");
	fclose(out);

	isDirty = false;
}

vector<CArchiveScanner::ModData> CArchiveScanner::GetPrimaryMods()
{
	vector<ModData> ret;

	for (map<string, ArchiveInfo>::iterator i = archiveInfo.begin(); i != archiveInfo.end(); ++i) {
		if (i->second.modData.name != "") {

			if (i->second.modData.modType != 1)
				continue;

			// Add the archive the mod is in as the first dependency
			ModData md = i->second.modData;
			md.dependencies.insert(md.dependencies.begin(), i->second.origName);
			ret.push_back(md);
		}
	}

	return ret;	
}

vector<string> CArchiveScanner::GetArchives(const string& root)
{
	vector<string> ret;

	string lcname = root;
	transform(lcname.begin(), lcname.end(), lcname.begin(), (int (*)(int))tolower);

	map<string, ArchiveInfo>::iterator aii = archiveInfo.find(lcname);
	if (aii == archiveInfo.end())
		return ret;

	ret.push_back(aii->second.path + "/" + aii->second.origName);

	if (aii->second.modData.name == "")
		return ret;

	// add depth-first
	for (vector<string>::iterator i = aii->second.modData.dependencies.begin(); i != aii->second.modData.dependencies.end(); ++i) {
		vector<string> dep = GetArchives(*i);
		for (vector<string>::iterator j = dep.begin(); j != dep.end(); ++j) {
			ret.push_back(*j);
		}
	}

	return ret;
}

vector<string> CArchiveScanner::GetMaps()
{
	vector<string> ret;

	for (map<string, ArchiveInfo>::iterator aii = archiveInfo.begin(); aii != archiveInfo.end(); ++aii) {
		for (vector<MapData>::iterator i = aii->second.mapData.begin(); i != aii->second.mapData.end(); ++i) {
			ret.push_back((*i).name);
		}
	}

	return ret;
}

vector<string> CArchiveScanner::GetArchivesForMap(const string& mapName)
{
	vector<string> ret;

	for (map<string, ArchiveInfo>::iterator aii = archiveInfo.begin(); aii != archiveInfo.end(); ++aii) {
		for (vector<MapData>::iterator i = aii->second.mapData.begin(); i != aii->second.mapData.end(); ++i) {
			if (mapName == (*i).name) {
				return GetArchives(aii->first);
			}
		}
	}

	return ret;
}

/*void CArchiveScanner::WriteData(FILE* out, const Data& data)
{
	fprintf(out, "\t\t\tName=%s;\n", data.name.c_str());
	fprintf(out, "\t\t\tNumDependencies=%d;\n", data.dependencies.size());
	int cur = 0;
	for (vector<string>::const_iterator i = data.dependencies.begin(); i != data.dependencies.end(); ++i) {
		fprintf(out, "\t\t\tDependency%d=%s;\n", cur, (*i).c_str());
		cur++;
	}
}*/